//
//								FREQAN
// Visualization program for seismic data stored in SEG-Y format 
// http://www.seg.org/documents/10161/77915/seg_y_rev1.pdf
// performs Fourier spectrum computation, time-frequency decomposition,
// and band-pass filtering
//
// Copyright 2011-2015 by Yuri Ganshin, CMI, University of Wyoming, Laramie.
// http://www.uwyo.edu/cmi/
//
// This program is free sofware released using the Fast Light Tool Kit (FLTK).
// For conditions of distribution and use, see the license agreement in 
//     http://www.fltk.org/COPYING.php
// As a general rule, "free software" means that the user have the freedom
// to run, copy, distribute, study, modify and improve the software.
//
// Please report all bugs and problems on the following page:
// http://www.uwyo.edu/cmi/
//     
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <cfloat>  // declares FLT_MAX
//
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Double_Window.H>
#include <FL/fl_draw.H>	
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Native_File_Chooser.H>
#include <FL/Fl_Int_Input.H>
// #include <FL/Fl_Slider.H>
#include <FL/Fl_Spinner.H>
#include <FL/Fl_Float_Input.H>
#include <FL/Fl_Input_Choice.H>
#include <FL/fl_show_colormap.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Box.H>
#include <FL/fl_message.H>
#include <FL/fl_ask.H> // for fl_alert() and fl_message()
#include <math.h>
#include <fftw3.h>
/**
The standard FFTW distribution works most efficiently and without any problem
for arrays whose size can be factored into small primes (2, 3, 5, and 7). 
It was noticed that for the array lengths with larger-size prime factors (17 and
above), the complex transforms might occasionally produce incorrect results in
the form of high-frequency noise.
**/
using namespace std;

/**********************   Spectra-Related   *******************/
#define Re(z,i) ((z)[2*(i)])
#define Im(z,i) ((z)[2*(i)+1])

#define FOLD		13		// Multiplicity of spectral stacks
#define TWIN		1000	// Max. time window [msec] for spectral analysis
#define PALSZ		11
#define TIME_BAR	60
#define HDR_BAR		24
#define RIM			 8		// RIM == border for canvas
#define MAX_VAL		100		// Maximum number of trace headers to plot
//
#ifndef FLT_MAX
#define FLT_MAX 1e+37
#endif
//
#ifndef M_PI
#define M_PI 3.14159265359
#endif
//
////////////////////////////////////////////////////////////////
/****               List of GLOBAL variables               ****/
FILE *in_fp = NULL;
char inpname[200];
int nsamp, dsamp, Ntr, convert_mode=1;
short measure;
int tr_sta,tr_fin,samp_sta,samp_fin,hdr_step=5;
int tr_vis=-1,samp_vis,wiggle_trace=0;
int numpix=3, trpick=1, sp1, sp2;
int hdr_val[MAX_VAL];
//
int canvas_w = 760, canvas_h =960;
int TF_w = 460, TF_h = 960;
int spec_w = 512, spec_h = 400;
//
int cmap=0, word_no=5;					// CDP for header plot
int slope_no=2;							// cosine slope type as default
unsigned int tr_size, samp_step=5;
float Norm, gain, elev_scal, coor_scal, *seis;
float *trace1, *trace2;
char *hdrs,c_word[4],slope_word[3],cgbuf[120];
char text_hdr[3201];
double f1=8.0, f2=15.0 ,f3=55.0, f4=80.0;	// four corner frequencies
//				SPECTROGRAM
int WIN,SR;
float fN;
double *bell[FOLD],norm[FOLD];
//
///////////////////  Color Palettes  ////////////////////
//				Normalized Spectral Amplitude
// array size = PALSZ*4,  palette size = PALSZ
float hot_pal[] = 
{   0,  10,  20,  30,  40,  50,  60,  70,  80,  90, 100,
  255, 255, 255, 255, 255, 255, 255, 190, 125,  63,   0,
  255, 255, 255, 190, 125,  63,   0,   0,   0,   0,   0,
  255, 100,  50,  25,   0,   0,   0,   0,   0,   0,   0};	
//
float YGB_pal[] =		// PALSZ = 11
{   0, 2.5, 5.0,  10,  25,  35,  60,  70,  80,  90, 100,
  255, 255,   0,   0, 255, 170,  85,   0,   0,   0,   0,
  255, 255, 175, 255, 255, 170,  85,   0,   0,   0,   0,
  175,   0,   0,   0, 255, 255, 255, 255, 170,  85,   0};	
//
float grey_pal[] =   // PALSZ = 11
{   0, 2.5, 5.0,  10,  25,  35,  60,  70,  80,  90, 100,
  255, 225, 200, 175, 150, 127, 105,  80,  55,  25,   0,
  255, 225, 200, 175, 150, 127, 105,  80,  55,  25,   0,
  255, 225, 200, 175, 150, 127, 105,  80,  55,  25,   0}; 

////////////////////////////////////////////////////////////////
/****            List of Function Prototypes               ****/

void read_file_hdr(const char *name);
void read_seis_dat();
void convert2(char* input);
void convert4(char* input);
void convert_hdrs(char* array);
float ibm2ieee(const float ibm);
float efloat(float value);
unsigned char to_ascii ( unsigned char c );
float conv3(float *dat,int NS1,int NS2,int index);
void pal_interp(float [], int, float, short []);

////////////////////////////////////////////////////////////////
/****         Global pointers to the GUI objects           ****/
////////////////////////////////////////////////////////////////
Fl_Double_Window *seis_window, *proc_window, *spec_window, *TF_window;
Fl_Window *opts_window, *hdr_window;
Fl_Int_Input	*tr_sta_field, *tr_fin_field, *trpick_field;
Fl_Int_Input	*t_sta_field,  *t_fin_field, *sp1_field, *sp2_field;
Fl_Float_Input	*f1_field, *f2_field, *f3_field, *f4_field, *gain_field;
Fl_Button *apply, *next_tr, *add_wiggle, *show_hdr, *end_this;
Fl_Button *apply_flt, *apply_spec, *apply_spgr;

////////////////////////////////////////////////////////////////
/****        List of Widgets and Drawing functions         ****/
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
/********             File-choosing MENU               ********/
////////////////////////////////////////////////////////////////
class File_menu : public Fl_Window {
	Fl_Native_File_Chooser *fc;
	
	void read_data(const char *filename)	// Process the chosen file
	{
		strcpy(inpname,filename);
		tr_vis=1;			// will wait while(tr_vis<0)
	}
	
	// Handle an 'Open' request from the menu
	static void open_cb(Fl_Widget *w, void *v) {
		File_menu *choice = (File_menu*)v;
		choice->fc->title("Open");
		choice->fc->type(Fl_Native_File_Chooser::BROWSE_FILE);
		switch ( choice->fc->show() ) {
			case -1: break;	// Error
			case  1: break; 	// Cancel
			default:		// choice->fc->show() returns 0
				choice->read_data(choice->fc->filename());
				// choice->deactivate();				
				break;
		}
	}

	static void quit_cb(Fl_Widget *w, void *v) {
		exit(0);
	}
	
public:
	File_menu() : Fl_Window(500,300,320,160,"Choose the file") {
		Fl_Menu_Bar *menu = new Fl_Menu_Bar(0,0,300,25);
		menu->add("&File/&Open",  FL_COMMAND+'o', open_cb, (void*)this);
		menu->add("&File/&Quit",  FL_COMMAND+'q', quit_cb);
		
		// Describe the chooser
		Fl_Box *box = new Fl_Box(20,45,w()-40,h()-65);
		box->color(FL_GRAY-1); 
		box->box(FL_FRAME_BOX);
		box->align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE|FL_ALIGN_WRAP);
		box->label("Choose seismic binary SEG-Y file\n\
with 4-byte floating point data format\n\
[ IBM, IEEE, or native byte order ]");
		box->labelsize(16);
		// Initialize the file chooser
		fc = new Fl_Native_File_Chooser();
		fc->filter("SEG-Y Binary\t*.sgy\n");			 
		end();
	}
};

//////////////////////////////////////////////////////////////////////
/**************        Drawing Seismic Data          ****************/
//   Double-buffered window eliminates flickering during updates
//////////////////////////////////////////////////////////////////////
class Draw_seis : public Fl_Double_Window 
{

public:
  	Draw_seis(int X,int Y,int W,int H) : Fl_Double_Window(X,Y,W,H) {}
	
	int handle(int event)
	{ 
		int xx = Fl::event_x(); 
		int yy = Fl::event_y();
		
		if(event==FL_PUSH)
		{
			// Returns 1 for FL_LEFT_MOUSE and 3 for FL_RIGHT_MOUSE			
			if(Fl::event_button()==FL_RIGHT_MOUSE)
			{
				int x_dim = w() - TIME_BAR;
				int y_dim = h() - HDR_BAR;
				
				trpick = tr_vis*(xx-TIME_BAR+x_dim/(2*tr_vis)+1)/x_dim;
				if(trpick<1) trpick=1; 
				if(trpick>tr_vis) trpick=tr_vis;
				sprintf(cgbuf,"%d",trpick);  trpick_field->value(cgbuf);
			}	
		}
		return Fl_Double_Window::handle(event); 
	}
	
	void draw(void)  // Function draw() within class Draw_seis
	{      /**      don't declare reserved 'x' and 'y' variables !!!    **/
		Fl_Double_Window::draw();
		unsigned i, j;
		int xx, xx1=0, yy, x_dim, y_dim, *hdr4;
		int val, tr1, tr2, sta_pix;
		short *hdr2;
		float yi, a1, a2;
		double dx,dtr;
		
		if(samp_vis<50) samp_step=5;		// Step for Time Labels
		else if(samp_vis<500) samp_step=25;
		else if(samp_vis<1000) samp_step=50;	
		else if(samp_vis<2000) samp_step=100;
		else if(samp_vis<5000) samp_step=250;
		else if(samp_vis<10000) samp_step=500;
		else samp_step=1000;
		
		hdr2=(short*)hdrs;	// To get trace ID (0=Unknown,1=Seismic,2=Dead)
		hdr4=(int*)hdrs;
		for(tr1=0,j=0;tr1<tr_vis;tr1+=hdr_step,j++) 
		{
			val = hdr4[tr1*60+word_no];
			if(word_no>=10 && word_no<=14) val=(int)((float)val*elev_scal);
			if(word_no>=18 && word_no<=46) val=(int)((float)val*coor_scal);
			if(j<MAX_VAL) hdr_val[j]=val;
		}			
		
		/**    Estimate X-shft from the number of numerals	  **/
		tr1=tr_vis/2;      // take middle trace as estimate
		val = hdr4[tr1*60+word_no];  // 60 = number_of_4-byte_words
		if(word_no>=10 && word_no<=14) val=(int)((float)val*elev_scal);
		if(word_no>=18 && word_no<=46) val=(int)((float)val*coor_scal);
		if(val<10) numpix=3;
		else if(val<100) numpix=7;
		else if(val<1000) numpix=11;
		else if(val<10000) numpix=15;		
		else if(val<100000) numpix=19;
		else if(val<1000000) numpix=23;
		else numpix=27;			//  numpix == X-shift in pixels
		
		x_dim = w() - TIME_BAR;  y_dim = h() - HDR_BAR;
		//cout << "Xdim = " << x_dim << ", Y_dim = " << y_dim << endl;	
		/////////////////////////////////////////////////////////////////

		fl_rectf(0,0,w(),h(),  0,  0,  0);			// Black bounding rectangle
		fl_rectf(1,1,w()-2,h()-2,230,230,230);		// Gray bounding rectangle
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(TIME_BAR,HDR_BAR,x_dim-1,y_dim-1,255,255,255);
		
		fl_color(FL_BLACK);	fl_font(0,14);	
		
		dtr = (double)x_dim/tr_vis;  //  decimal  pixels/trace
		xx1 = sta_pix = TIME_BAR + (int)(dtr/2.0) + 1;
		
		///////////////////////////////////////////////////////////////////////
		//  Variable Area Display utilizing cubic convolution for resampling //
		//  and linear interpolataion for filling inter-trace space			 //
		///////////////////////////////////////////////////////////////////////
		for(tr1=0;tr1<tr_vis;tr1++)
		{										
			i = tr1*x_dim/tr_vis;
			xx = i + sta_pix;
			tr2 = tr1+1; 
			dx = fabs((double)i - dtr*tr1)/dtr;	//   0.0 <= dx <= 1.0
			
			if(hdr2[tr1*120+14]==2) for(i=0;i<nsamp;i++) trace1[i]=0.0; // avoid NAN's
			else memcpy(trace1, seis+tr1*nsamp, nsamp * sizeof(float));
			
			if(hdr2[tr2*120+14]==2) for(i=0;i<nsamp;i++) trace2[i]=0.0; // avoid NAN's
			else memcpy(trace2, seis+tr2*nsamp, nsamp * sizeof(float));			
			
			for(j=0;j<y_dim;j++)  // possible: j+=3 for fl_yxline() 
			{
				yy = j+HDR_BAR;	// yy = 24 (j==0)
				// Resample data for visualization using Cubic Convolution Method 
				a1=conv3(trace1+samp_sta,samp_vis,y_dim,j);			
				a2=conv3(trace2+samp_sta,samp_vis,y_dim,j);
				yi = a1+(a2-a1)*dx; // Inter-trace linear interpol: a1<=yi<=a2
				val = (int)((yi*gain*15)/Norm);
				if(val>0)fl_xyline(xx, yy, xx+val); // horiz line for VA display
				if(wiggle_trace) if(j>0) fl_line(xx+val,yy,xx1,yy-1); // overlay
				xx1 = xx+val;						
			}								
		}		
		
		for(tr1=0,j=0;tr1<tr_vis;tr1+=hdr_step,j++)	////  Draw Trace Headers
		{
			i = tr1*x_dim/tr_vis;
			xx = i + sta_pix;
			fl_yxline(xx, HDR_BAR, HDR_BAR-6);		 
			if(j<MAX_VAL) sprintf(cgbuf,"%d",hdr_val[j]); 
			else sprintf(cgbuf,"%d",99999);
			fl_draw(cgbuf,xx-numpix,HDR_BAR-10);	
		}											////  Draw Trace Headers
		
		/////////////////////////////////////////////////  Draw Time labels
		sprintf(cgbuf,"Time ");  fl_draw(cgbuf,10,HDR_BAR+8);
		sprintf(cgbuf,"[ ms ]");  fl_draw(cgbuf,10,HDR_BAR+26);	
		for(i=0;i<samp_vis;i+=samp_step)
		{
			yy = y_dim*i/samp_vis + HDR_BAR;
			if(i>0) 
			{
				val=(samp_sta+i)*dsamp/1000;
				sprintf(cgbuf,"%5d",val); fl_draw(cgbuf,10,yy+4); // text string
				fl_xyline(TIME_BAR-10,yy,TIME_BAR+x_dim);  // horiz line			
			}
		}
		
	} 	// void draw(void)	
	
} *seismic; 

//
/////////////////////////////////////////////////////////////////////
/**************       Drawing Processed Data        ****************/
/////////////////////////////////////////////////////////////////////
class ProcCanvas : public Fl_Widget 
{
public:
	ProcCanvas(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}

	void draw()
	{
		unsigned i, j, tr, pos;
		int xx, xx1=0, yy, x_dim, y_dim;
		int val, t, tr1, tr2, sta_pix;
		short *hdr2;
		float yi, a1, a2, *proc;
		double dx,dtr;
		// Fast Fourier Transform related:
		fftw_complex *in,*out;
		fftw_plan fwd,bck;
		double *tmp1,*tmp2,*resp,f,df12,df34;
					
	
		x_dim = w() - TIME_BAR;  y_dim = h() - HDR_BAR;
		/////////////////////////      !!!!!!     //////////////////////////////
		fl_rectf(x()-1,y()-1,w()+2,h()+2,  0,  0,  0);	// Black bounding rectangle
		fl_rectf(x(),  y(),  w(),  h(),  230,230,230);	// Gray bounding rectangle
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(x()+TIME_BAR,y()+HDR_BAR,x_dim,y_dim,255,255,255);
		fl_color(FL_BLACK);	fl_font(0,14);	 // Helvetica regular, 14 pixel
		dtr = (double)x_dim/tr_vis;  //  decimal  pixels/trace
		xx1 = sta_pix = TIME_BAR + (int)(dtr/2.0) + 1;
		hdr2=(short*)hdrs;	// To get trace ID (0=Unknown,1=Seismic,2=Dead)
		
		///////////////////////////////////////////////////////////////////////
		////     Processing    Processing    Processing    Processing	   ////
		///////////////////////////////////////////////////////////////////////

		//  sizeof(fftw_complex) = 16 bytes
		in = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * nsamp);
		out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * nsamp);
		// Forward and backward in-place transforms:		
		fwd = fftw_plan_dft_1d(nsamp,in,in,-1,FFTW_ESTIMATE); 
		bck = fftw_plan_dft_1d(nsamp,out,out,1,FFTW_ESTIMATE);
		
		tmp1=(double *)in;	// To translate complex-number type into double		
		tmp2=(double *)out;
			
		resp = new (nothrow) double[1 + nsamp/2];
		proc = new (nothrow) float[(tr_vis+1) * nsamp];
		if(proc == 0)	// Do not damage original seis[] data
		{ 
			fl_message("ERROR: NO memory to process data!\n\
						  Close the window to exit...");	   exit(1);
		}
		
		////				Trapezodal-shape FREQUENCY RESPONSE function
//cout << "F1=" << f1 << ", F2=" << f2 << ", F3=" << f3 << ", F4=" << f4 << endl;
		
		df12=f2-f1;	df34=f4-f3;
		for(i=0;i<=nsamp/2;i++)	// scan up to Nyquist Frequency
		{
			f = 1000000.0*i/(dsamp*nsamp);		// frequency, Hertz
			resp[i] = 0.0;
			if(slope_no==1)
			{
				if(f>f1 && f<f2) resp[i] = (f-f1)/df12; // 0->1
				if(f>=f2 && f<=f3) resp[i] = 1.0;
				if(f>f3 && f<f4) resp[i] = 1.0 - (f-f3)/df34; // 1->0			
			}
			else
			{
				if(f>f1 && f<f2) resp[i]=(1.0-cos(M_PI*(f-f1)/df12))/2.0; // 0->1
				if(f>=f2 && f<=f3) resp[i]=1.0;
				if(f>f3 && f<f4) resp[i]=(1.0+cos(M_PI*(f-f3)/df34))/2.0; // 1->0
			}
			// cout  << f << " " << resp[i] << endl;	
		}		
		
		for(tr=0;tr<tr_vis;tr++)	////    FILTERING    FILTERING    FILTERING
		{
			if(hdr2[tr*120+14]==2) for(i=0;i<nsamp;i++) trace1[i]=0.0; // avoid NAN's
			else memcpy(trace1, seis+tr*nsamp, nsamp * sizeof(float));
		
			//  Read seismic data into complex array and transform it
			for(i=0;i<nsamp;i++) { Re(tmp1,i)=trace1[i]; Im(tmp1,i)=0.0; }
			fftw_execute(fwd);		// Forward in-place transform with in[]
			
			// Convolve in[] with filtr response resp[] and output in out[]
			for(i=0;i<=nsamp/2;i++)	// scan up to Nyquist Frequency
			{				
				Re(tmp2,i) = Re(tmp1,i)*resp[i] - Im(tmp1,i)*resp[i];
				Im(tmp2,i) = Re(tmp1,i)*resp[i] + Im(tmp1,i)*resp[i];
				if(i==0) continue;	// do not duplicate zero-frequency component
				if(i==nsamp/2) break;// do not duplicate Nyqusist frequency
				j=nsamp-i;			//  Backwards order for 'negative' frequencies
				Re(tmp2,j) = Re(tmp1,j)*resp[i] - Im(tmp1,j)*resp[i];
				Im(tmp2,j) = Re(tmp1,j)*resp[i] + Im(tmp1,j)*resp[i];			
			}
			
			fftw_execute(bck);		// Backward in-place transform with out[]
			for(i=0;i<nsamp;i++)	// Fill-in proc[] and normalize with 'nsamp'
				proc[tr*nsamp+i]=(float)(Re(tmp2,i)/nsamp);
		}

		///////////////////////////////////////////////////////////////////////
		//        Visualization      Visualization     Visualization	  	 //
		///////////////////////////////////////////////////////////////////////
		//  Variable Area Display utilizing cubic convolution for resampling //
		//  and linear interpolataion for filling inter-trace space			 //
		///////////////////////////////////////////////////////////////////////

		for(tr1=0;tr1<tr_vis;tr1++)
		{		
			i = tr1*x_dim/tr_vis;
			xx = x() + i + sta_pix;
			dx = fabs((double)i - dtr*tr1)/dtr;	//   0.0 <= dx <= 1.0			
			tr2 = tr1+1;
			
			memcpy(trace1, proc+tr1*nsamp, nsamp * sizeof(float));
			memcpy(trace2, proc+tr2*nsamp, nsamp * sizeof(float));			
 
			for(j=0;j<y_dim;j++)  // possible: j+=3 for fl_yxline() 
			{
				yy = y()+j+HDR_BAR;	// yy = 24 (j==0)
				a1=conv3(trace1+samp_sta,samp_vis,y_dim,j);
				a2=conv3(trace2+samp_sta,samp_vis,y_dim,j);
				yi = a1+(a2-a1)*dx; // Inter-trace linear interpol: a1<=yi<=a2
				val = (int)((yi*gain*15)/Norm);
				if(val>0)fl_xyline(xx, yy, xx+val); // horiz line for VA display
				if(wiggle_trace) if(j>0) fl_line(xx+val,yy,xx1,yy-1); // overlay
				xx1 = xx+val;
			}								
		}		
		
		for(tr1=0,j=0;tr1<tr_vis;tr1+=hdr_step,j++)	////  Draw Trace Headers 
		{
			i = tr1*x_dim/tr_vis;
			xx = x() + i + sta_pix;
			fl_yxline(xx, y()+HDR_BAR, y()+HDR_BAR-6);		 
			if(j<MAX_VAL) sprintf(cgbuf,"%d",hdr_val[j]); 
			else sprintf(cgbuf,"%d",99999);
			fl_draw(cgbuf,xx-numpix,y()+HDR_BAR-10);	
		}			
		
		xx = x()+10;  yy = y()+HDR_BAR;					////  Draw Time labels
		sprintf(cgbuf,"Time ");  fl_draw(cgbuf,xx,yy+8);
		sprintf(cgbuf,"[ ms ]"); fl_draw(cgbuf,xx,yy+26);
				
		xx = x()+TIME_BAR;
		for(i=0;i<samp_vis;i+=samp_step)
		{
			yy = y_dim*i/samp_vis + y() + HDR_BAR;
			if(i>0) 
			{
				val=(samp_sta+i)*dsamp/1000;
				sprintf(cgbuf,"%5d",val); fl_draw(cgbuf,x()+10,yy+4);
				fl_xyline(xx-10, yy, xx+x_dim);	  // horiz line		
			}
		}
		
		delete[] resp;  delete[] proc;
		fftw_destroy_plan(fwd);   fftw_destroy_plan(bck); 
		fftw_free(in); fftw_free(out);
		
	}	// void draw()
	
} *proc_draw;

/////////////////////////////////////////////////////////////////////
///////////         NON-SCROLLABLE SPECTRUM       ///////////////////
/////////////////////////////////////////////////////////////////////
class SpecCanvas : public Fl_Widget 
{
	
public:
	SpecCanvas(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}
	
	void draw()
	{
		int i,j,tr,f,NN,SR,xx,yy,sta,end,vali,Nend;
		int f_mark,f_inc=10;  // 10 Hz - increment for frequency marks
		short *hdr2;
		// Fast Fourier Transform related:
		fftw_complex *in;
		fftw_plan fwd;			
		double val,sp_norm,*tmp,*taper,*spec;
		
		///////////////////////////////////////////////////////////////	
		fl_rectf(x()-1,y()-1,w()+2,h()+2,  0,  0,  0);	// Black rectangle
		fl_rectf(x(),  y(),  w(),  h(),  255,255,255);	// White WORK AREA

		///////////////////////////////////////////////////////////////
		////////     Spectrum-related  memory allocation     //////////
		///////////////////////////////////////////////////////////////
		
		hdr2=(short*)hdrs;	// To get trace ID (0=Unknown,1=Seismic,2=Dead)
				
		NN = samp_vis;  SR = 1000000/dsamp; // samples/second
		if(SR>=500) Nend = NN/4;	// Spectral line corresponding fN/2
		else Nend = NN/2;			// Spectral line corresponding fN

		spec = (double*)malloc(NN*sizeof(double));
		taper =(double*)malloc(NN*sizeof(double));		
		
		in = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * NN);
		if(!in)
		{ fl_message("ERROR: NO memory to accomodate spectrum!\n\
						 Close the window to exit...");	   exit(1); }
		fwd = fftw_plan_dft_1d(NN,in,in,-1,FFTW_ESTIMATE);  // Forward
		tmp=(double *)in;	// To translate complex-number type into double		
		
		sta = NN/20;  end = 19*NN/20;		
		for(i=0;i<NN;i++)		////  Tapering function with cosine-slopes 
		{
			taper[i] = 1.0;
			if(i<sta) taper[i]=(1.0+cos(M_PI*(i-sta)/sta))/2.0; // 0 -> 1
			if(i>end) taper[i]=(1.0+cos(M_PI*(i-end)/sta))/2.0; // 1 -> 0
		}				
		
		/////////////////////////////////////////////////////////////
		//////// 		Calculate Cumulative Spectra		/////////

		for(i=0;i<NN;i++) spec[i] = 0.0;	// Initialization with zeroes
		
		for(tr=0;tr<tr_vis;tr++)
		{
			if(hdr2[tr*120+14]==2) continue;
			if(tr+tr_sta<sp1) continue;
			if(tr+tr_sta>sp2) break;
			
			//  Read seismic data into complex array and transform it			
			for(i=0,j=samp_sta;i<NN;i++,j++) 
			{ Re(tmp,i)=taper[i]*seis[tr*nsamp+j]; Im(tmp,i)=0.0; }
			fftw_execute(fwd);	// Forward in-place transform

			//  Cumulative Amplitude Spectrum for positive frequencies
			for(i=1;i<=NN/2;i++)		//  up to Nyquist		
				spec[i] += sqrt(Re(tmp,i)*Re(tmp,i)+Im(tmp,i)*Im(tmp,i))/NN;
		}

		sp_norm=0.0;		// Normalize with maximum (or average)
		sta = Nend/10;  end  = 9*Nend/10;
		for(i=sta;i<end;i++) { val = spec[i]; if(val>sp_norm) sp_norm=val; }
		
		/////////////////////////////////////////////////////////////
		//////// 	Plot Cumulative Amplitude Spectrum		/////////		
		
		fl_line_style(0,1);  fl_font(1,12);  // Helvetica, 12 pixel
		f_mark = f_inc; 
		for(j=0;j<spec_w;j++)	// Scan for window pixels
		{
			xx = x() + j;		// X-coordinate in pixels
			i = j*Nend/spec_w;	// corresponding spectral line, up to Nend
			f = SR*i/NN;		// corresponding frequency, Hz 
			
			if(f>=f_mark)		// frequency marks [Hz]
			{
				fl_color(FL_GRAY);
				fl_yxline(xx, RIM, RIM+h());		// Gray vert line
				sprintf(cgbuf,"%d",f_mark);  fl_color(FL_BLACK);
				fl_yxline(xx, RIM+h(), RIM+h()+4);	// Black vert line
				fl_draw(cgbuf,xx-6,RIM+h()+20); f_mark += f_inc;
			}						
			vali = (int)(spec[i]*spec_h/sp_norm);

			yy = h() + RIM - vali;  if(yy<RIM) yy=RIM;		
			fl_color(FL_BLUE);  fl_yxline(xx, h()+RIM, yy);	// vert line
		}
		///////////////////////////////////////////////////////////////
		
		fl_font(1,14);			// Helvetica-Bold, 14 pixel
		fl_color(FL_BLACK);		
		sprintf(cgbuf,"Frequency [Hz]"); 
		fl_draw(cgbuf,x()-50+w()/2,y()+h()+38);
		
		fl_font(0,12);
		sprintf(cgbuf,"Cumulative Amplitude Spectrum");
		fl_draw(cgbuf,x()-180+w(),y()+15);			
		sprintf(cgbuf,"for traces from %d to %d",sp1,sp2);		
		fl_draw(cgbuf,x()-180+w(),y()+30);
		sprintf(cgbuf,"and time from %d to %d ms",
				  samp_sta*dsamp/1000,samp_fin*dsamp/1000); 
		fl_draw(cgbuf,x()-180+w(),y()+45);						
		
		////////////////////////////////////////////////////////////////////		
		
		fftw_destroy_plan(fwd);
		fftw_free(in);	free(taper); free(spec);

	}	// void draw()
	
} *spec_draw;

/////////////////////////////////////////////////////////////////////
///////////         NON-SCROLLABLE SPECTROGRAM       ////////////////
/////////////////////////////////////////////////////////////////////
class TFCanvas : public Fl_Widget 
{
public:
	TFCanvas(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}
	
	void draw()
	{
		int i,j,k,samp,N,Nsp,t,freq;
		int ii, jj, xx, yy;
		short rgb[3];		
		float *winbuf,sp;
		double val,std,max;	
		double *spec,*tmp;
					
		// Fast Fourier Transform related:
		fftw_complex *arr;
		fftw_plan fwd;	

		int x_dim = w() - TIME_BAR;
		int y_dim = h() - HDR_BAR;				

		fl_rectf(x()-1,y()-1,w()+2,h()+2,0,0,0);	// Black bounding rectangle 
		fl_rectf(x(),y(),w(),h(),230,230,230);		// Gray bounding rectangle
		///////    Finally, clean the WORK AREA with WHITE
		fl_rectf(x()+TIME_BAR,y()+HDR_BAR,x_dim,y_dim,255,255,255);
		
		/////////////////////////////////////////////////////////////////
		/////////////      Definitions and Memory       /////////////////
				
		N = WIN/2;			// Spectral line number that corresponds fN
		Nsp = 100;			// 100 spect. lines == 100 Hz for TWIN=1000
		
		arr = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * WIN);
		fwd = fftw_plan_dft_1d(WIN,arr,arr,-1,FFTW_ESTIMATE); // in-place
		tmp=(double *)arr; // To translate complex-number type into double
		
		winbuf = (float*) malloc(WIN*sizeof(float));
		spec = (double*)malloc(samp_vis*N*sizeof(double));
		
		/////////////////////////////////////////////////////////////////
		/////////////      SPECTRA CALCULATION          /////////////////
		
		for(k=0,samp=samp_sta; k<samp_vis; k++,samp++)
		{
			for(i=0;i<WIN;i++)
			{
				winbuf[i] = 0.0;
				j = samp + i - N;
				if(j>=0 && j<nsamp) winbuf[i]=seis[trpick*nsamp+j];
			}
			
			for(i=0;i<N;i++) spec[k*N+i] = 0.0;			
			for(j=0;j<FOLD;j++) // Balancing uncertainty and resolution
			{
				for(i=0;i<WIN;i++)
				{ Im(tmp,i) = 0.0;  Re(tmp,i) = bell[j][i]*winbuf[i]; }
				
				fftw_execute(fwd);
				
				for(i=0;i<N;i++) // Stack spectra for balanced T-F resolution
					spec[k*N+i]+=sqrt(Re(tmp,i)*Re(tmp,i)+Im(tmp,i)*Im(tmp,i))/norm[j];
			}
		}		// for(samp...)
		
		///////    SPECTRA NORMALIZATION to absolute maximum     //////////
		max = 0.0;
		for(k=0;k<samp_vis;k++)
			for(i=0;i<N;i++) { val=spec[k*N+i];	if(val>max) max=val; }
		if(max==0.0) max=1.0;	

		/////////////////////////////////////////////////////////////////
		/////////////      SPECTRA VISUALIZATION        /////////////////
	
		for(jj=0; jj<y_dim; jj++)			//   Cycle pixels in Y-direction
		{
			k = jj*samp_vis/y_dim;			// spec[] index			
			for(i=0;i<Nsp;i++) winbuf[i] = (float)(100.0*spec[k*N+i]/max); 

			yy = y() + jj + HDR_BAR;		// Y-coordinate for drawings
			
			for(ii=0; ii<x_dim; ii++)		//	Cycle pixels in X-direction
			{			
				sp=conv3(winbuf,Nsp,x_dim,ii);	
				pal_interp(YGB_pal,PALSZ,sp,rgb);
				
				xx = x() + ii + TIME_BAR;	// X-coordinate for drawings
				fl_rectf(xx,yy,1,1,rgb[0],rgb[1],rgb[2]);
			}			
		}
		
		/////////////////////////////				Draw Frequency labels		
		fl_color(FL_BLACK);	fl_font(1,14);		// Helvetica bold, 14 pixel
		sprintf(cgbuf,"Frequency, Hz");
		fl_draw(cgbuf,x()+w()-100,y()+15);		
		
		freq=0;	// starting label
		for(ii=0; ii<(x_dim-1); ii++)	/////////	Cycle pixels in X-direction		
		{
			k = (ii+1)*Nsp/x_dim;	// current frequency values in Hz
			
			if(k>=freq)
			{	
				xx = x() + ii + TIME_BAR;
				sprintf(cgbuf,"%3d",freq);
				if(xx<x()+w()-90) fl_draw(cgbuf,xx-12,y()+15);	// text string
				fl_yxline(xx,y()+HDR_BAR-4,y()+HDR_BAR+y_dim); // vertical line
				freq += 10;				// Increment for frequency labels
			}
		}
		
		/////////////////////////////				Draw Time labels		
		sprintf(cgbuf,"Time ");   fl_draw(cgbuf,x()+8,y()+HDR_BAR+12);
		sprintf(cgbuf,"[ ms ]");  fl_draw(cgbuf,x()+8,y()+HDR_BAR+30);
		
		for(jj=0;jj<samp_vis;jj+=samp_step)
		{
			yy = y() + y_dim*jj/samp_vis + HDR_BAR;
			t=(samp_sta+jj)*dsamp/1000;
			sprintf(cgbuf,"%5d",t);
			if(jj>0) fl_draw(cgbuf,15,yy+4);		// draw text string
			fl_xyline(x()+TIME_BAR-6, yy, x()+TIME_BAR+x_dim);
		}
			
		/////////////      Free  Memory       /////////////////					
		fftw_destroy_plan(fwd);
		fftw_free(arr);	free(winbuf);  free(spec);					
		
	}	// void draw()
	
} *TF_draw;

////////////////////////////////////////////////////////////////
/************    List of Call Back functions    ***************/
////////////////////////////////////////////////////////////////

void input_tr_sta(Fl_Widget *, void *) {
	tr_sta = atoi(tr_sta_field->value());
	sp1=tr_sta;
	sprintf(cgbuf,"%d",sp1);  sp1_field->value(cgbuf);
}

void input_tr_fin(Fl_Widget *, void *) {
	tr_fin = atoi(tr_fin_field->value());
	sp2=tr_fin;
	sprintf(cgbuf,"%d",sp2);  sp2_field->value(cgbuf);
}

void input_t_sta(Fl_Widget *, void *) {
	samp_sta = atoi(t_sta_field->value())*1000/dsamp;
}

void input_t_fin(Fl_Widget *, void *) {
	samp_fin = atoi(t_fin_field->value())*1000/dsamp;
}

void input_gain(Fl_Widget *, void *) {
	gain = atof(gain_field->value());
}

void spinner_cb(Fl_Widget *w, void *) {
   Fl_Spinner *s = (Fl_Spinner *)w;	
   hdr_step = (int)s->value();
	// printf("Spinner Value = %d\n",hdr_step);
}

void input_choice_cb(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_hdr=(Fl_Input_Choice *)data;
	
	strncpy(c_word,in_hdr->value(),3);
	sscanf(c_word,"%d", &word_no);
}

void slope_choice(Fl_Widget*,void*data) {
	Fl_Input_Choice *slope=(Fl_Input_Choice *)data;
	
	strncpy(slope_word,slope->value(),2);
	sscanf(slope_word,"%d", &slope_no);
}

void show_hdr_cb(Fl_Widget *, void *) 
{
   // fl_message(text_hdr); -> accomdates only 12 rows out of 40
	hdr_window = new Fl_Window(676,676,"File's Textual header");

	Fl_Box* txt = new Fl_Box(FL_FRAME_BOX,8,8,660,660,text_hdr);
	txt->color(FL_WHITE);	
	txt->labelfont(0);
	txt->labelsize(14);
	
	hdr_window->resizable(txt); 
	hdr_window->end();
	hdr_window->show();
}

void input_f1(Fl_Widget *, void *) {
	f1 = atof(f1_field->value());
}

void input_f2(Fl_Widget *, void *) {
	f2 = atof(f2_field->value());
}

void input_f3(Fl_Widget *, void *) {
	f3 = atof(f3_field->value());
}

void input_f4(Fl_Widget *, void *) {
	f4 = atof(f4_field->value());
}

void input_sp1(Fl_Widget *, void *) {
	sp1 = atoi(sp1_field->value());
}

void input_sp2(Fl_Widget *, void *) {
	sp2 = atoi(sp2_field->value());
}

void input_trpick(Fl_Widget *, void *) {
	trpick = atoi(trpick_field->value());
}

void exit_b(Fl_Widget *, void *) {
	exit(0);
}

// Redraw display with updated data
void doRedraw(Fl_Widget *, void*)
{
	if(tr_fin>Ntr) tr_fin=Ntr;	
	if(tr_sta>=tr_fin) tr_sta=1; 
	tr_vis = tr_fin-tr_sta+1; // if trace count starts from '1'

	if(samp_fin>nsamp) samp_fin=nsamp;
	if(samp_sta>=samp_fin) samp_sta=0;
	samp_vis = samp_fin-samp_sta;
		
	read_seis_dat();
	seismic->redraw();
}

void doNext(Fl_Widget*, void*)
{
	if(tr_fin<Ntr)
	{
		tr_fin += tr_vis;			// tr_vis remains unchanged
		if(tr_fin>Ntr) tr_fin=Ntr;
		tr_sta = tr_fin-tr_vis+1;	// from tr_vis formula

		sprintf(cgbuf,"%d",tr_sta);  tr_sta_field->value(cgbuf);
		sprintf(cgbuf,"%d",tr_fin);  tr_fin_field->value(cgbuf);
		
		sp1=tr_sta; sp2=tr_fin;
		sprintf(cgbuf,"%d",sp1);  sp1_field->value(cgbuf);
		sprintf(cgbuf,"%d",sp2);  sp2_field->value(cgbuf);
			
		read_seis_dat();
		seismic->redraw();
	}
}

void do_wiggle(Fl_Widget*, void*)
{
	if(wiggle_trace==0) wiggle_trace=1;
	else wiggle_trace=0; 
	seismic->redraw();
}

////////////////////////////////////////////////////////////////////////
/////////////////    Window with FILTERED data      ////////////////////	
////////////////////////////////////////////////////////////////////////
void Process(Fl_Widget *, void*)
{  

	if(slope_no>2 || slope_no<1)
	{			
		fl_message("WARNING: Undefined slope type [# %d]\n\
Close the window to continue...",slope_no);
		return;
	}
		
	sprintf(cgbuf,"Band-pass cutoff frequencies: %.0lf-%.0lf-%.0lf-%.0lf Hz",
	                                f1,f2,f3,f4);
	proc_window = new Fl_Double_Window(canvas_w+RIM*2,canvas_h+RIM*2,cgbuf);
			
	proc_draw = new ProcCanvas(RIM,RIM,canvas_w,canvas_h);
	//seismic->redraw();	

	proc_window->end();
	proc_window->resizable(proc_window); // Critical !!!
	proc_window->show();
}

////////////////////////////////////////////////////////////////////////
/////////////////       Window with SPECTRUM        ////////////////////	
////////////////////////////////////////////////////////////////////////

void Spectrum(Fl_Widget *, void*)
{  
	
	if(sp2>tr_fin || sp1<tr_sta || sp1>sp2)
	{			
		fl_message("WARNING: traces selected for spectral analysis\n\
 should be within the range of displayed traces.\n\
 Close the window to continue...");
		return;
	}
	
	sprintf(cgbuf,"Amplitude Spectrum for traces %d to %d",sp1,sp2);
	spec_window = new Fl_Double_Window(spec_w+RIM*2,spec_h+RIM+50,cgbuf);
	
	spec_draw = new SpecCanvas(RIM,RIM,spec_w,spec_h);

	spec_window->end();
	spec_window->show();
}

////////////////////////////////////////////////////////////////////////
////////////////       Window with SPECTROGRAM        //////////////////	
////////////////////////////////////////////////////////////////////////
void SPECgram(Fl_Widget *, void*)
{  
	trpick--;
	if(trpick<0) trpick=0; 
	if(trpick>=tr_vis) trpick=tr_vis-1;

	sprintf(cgbuf,"Time-Frequency Representation for trace %d",trpick+1);	
	TF_window = new Fl_Double_Window(TF_w+RIM*2,TF_h+RIM*2,cgbuf);
	TF_draw = new TFCanvas(RIM,RIM,TF_w,TF_h);  // X,Y,W,H
	
	TF_window->end();
	TF_window->show();	
}

/////////////////////////////////////////////////////////////////////
/****   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   ****/
/////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
	int i,tr;
	double t,val;
	char cbuf1[90],cbuf2[90],cbuf3[90],cbuf4[90];
	char cf1[20],cf2[20],cf3[20],cf4[20];
	char cbuf[90],plot_hdr[120];	
	///////////////////////////////////////////////////////////
	
	Fl::scheme(NULL); //  "plastic" for Aqua and "gtk+"
	// NULL use the scheme defined in the FLTK_SCHEME env.var. 
	
	if(argc == 2) strcpy(inpname,argv[1]);		// Do not use terminal
	else		// if you want to create an Apple  *.app bundle application
	{
		File_menu *choice = new File_menu();
		choice->show();
		choice->end(); // no more widgets will be added to the window
		while(tr_vis<0) Fl::wait();
		choice->hide();
	}

	read_file_hdr(inpname);
	read_seis_dat();			// into new seis[]

	///////////////////////////////////////////////////////////
	strcpy(plot_hdr,"Seismic Data");
	seis_window = new Fl_Double_Window(canvas_w+RIM*2,canvas_h+RIM*2,plot_hdr);
	
	seismic = new Draw_seis(RIM,RIM,canvas_w,canvas_h);
	// seismic->end();	
	
	seis_window->end();
	seis_window->resizable(seis_window); // Critical !!!
	seis_window->show();

	///////////////////////////////////////////////////////////
		
	opts_window = new Fl_Window(1050,50,420,972,"Display Manager");
	// The follwing widgets belong to this window	

	Fl_Box* label_o = new Fl_Box(FL_FRAME_BOX,20,10,380,24,
						"==========   Trace Display Parameters   ==========");
	label_o->labelfont(0);  label_o->labelcolor(7);
	label_o->color(36);		label_o->labelsize(14);
	//label_o->labeltype(FL_ENGRAVED_LABEL);

	Fl_Box* label_tr = new Fl_Box(80,30,140,50,"Range of displayed traces :");
	label_tr->labelfont(FL_HELVETICA_BOLD);
	label_tr->labelsize(14);
	
	tr_sta_field = new Fl_Int_Input(80, 70, 70, 25, " from ");
	sprintf(cbuf,"%d",tr_sta);  tr_sta_field->value(cbuf);
	tr_sta_field->callback(input_tr_sta);
	tr_fin_field = new Fl_Int_Input(180, 70, 70, 25, " to ");
	sprintf(cbuf,"%d",tr_fin);  tr_fin_field->value(cbuf);
	tr_fin_field->callback(input_tr_fin);
	
	next_tr = new Fl_Button(285, 70, 100, 25, " Next traces");
	next_tr->color(62);
	next_tr->labelfont(1);	
	next_tr->callback(doNext);

	sprintf(cbuf1,"Total number of traces = %d",Ntr);	
	Fl_Box* comment1 = new Fl_Box(80,82,150,50,cbuf1);
	comment1->labelfont(FL_HELVETICA);
	comment1->labelsize(14);		
		
	Fl_Box* label_t = new Fl_Box(80,110,140,50,"Time range [ms] to display:");
	label_t->labelfont(FL_HELVETICA_BOLD);
	label_t->labelsize(14);	
	
	t_sta_field = new Fl_Int_Input(80, 150, 70, 25, " from ");
	sprintf(cbuf,"%d",samp_sta*dsamp/1000); t_sta_field->value(cbuf); 
	t_sta_field->callback(input_t_sta);
	t_fin_field = new Fl_Int_Input(180, 150, 70, 25, " to ");
	sprintf(cbuf,"%d",samp_fin*dsamp/1000); t_fin_field->value(cbuf); 
	t_fin_field->callback(input_t_fin);
	
	gain_field = new Fl_Float_Input(310, 150, 70, 25, "Gain: ");
	sprintf(cbuf,"%5.2f",gain);  gain_field->value(cbuf);
	gain_field->labelsize(14);
	gain_field->labelfont(FL_HELVETICA_BOLD);	
	gain_field->callback(input_gain);
	
	sprintf(cbuf4,"Total trace length = %d msec",nsamp*dsamp/1000);	
	Fl_Box* comment4 = new Fl_Box(80,162,150,50,cbuf4);
	comment4->labelfont(FL_HELVETICA);
	comment4->labelsize(14);	

	show_hdr = new Fl_Button(200,200,180,25,"Show file's text header");
	show_hdr->color(62);									
	show_hdr->labelcolor(FL_BLACK);
	show_hdr->labelfont(3);	
	show_hdr->labelsize(14);	
	show_hdr->callback(show_hdr_cb);	
	
	Fl_Box* label_hdr = new Fl_Box(50,225,300,54,
									"Select header plot parameter\n\
from the list of SEGY trace headers:");
	label_hdr->labelfont(FL_HELVETICA_BOLD);
	label_hdr->labelsize(14);
	Fl_Input_Choice in_hdr(50,275,300,30,"");
	in_hdr.callback(input_choice_cb, (void*)&in_hdr);
	
	in_hdr.add("  2 - Field record number");
	in_hdr.add("  3 - Trace number within Field record");
	in_hdr.add("  4 - Energy source point number"); 
	in_hdr.add("  5 - Ensemble number (CDP, CRT, ...)");
	in_hdr.add("  6 - Trace number within the ensemble");
	in_hdr.add("  9 - Source-receiver offset");
	in_hdr.add(" 10 - Receiver group elevation");  
	in_hdr.add(" 11 - Surface elevation at source");
	in_hdr.add(" 13 - Datum elevation at receiver");
	in_hdr.add(" 14 - Datum elevation at source");
	in_hdr.add(" 18 - Source coordinate X");
	in_hdr.add(" 19 - Source coordinate Y");
	in_hdr.add(" 20 - Group coordinate X");
	in_hdr.add(" 21 - Group coordinate Y");
	in_hdr.add(" 45 - ensemble (CMP) coordinate X");
	in_hdr.add(" 46 - ensemble (CMP) coordinate Y");
	in_hdr.add(" 47 - In-line number for 3-D data");
	in_hdr.add(" 48 - Cross-line number for 3-D data");
	in_hdr.add(" 49 - Shotpoint number (for 2-D data)");		
	in_hdr.value(3);  // CDP - default header plot parameter
	
	Fl_Box* comment = new Fl_Box(50,300,300,50,
					"the leading number indicates 4-byte integer word,\n\
starting from 0, within the 240-byte trace header");
	comment->labelfont(FL_HELVETICA);
	comment->labelsize(14);
	
	Fl_Spinner* spin = new Fl_Spinner(250, 350, 70, 30, 
										"Header spacing, traces ");
	spin->labelsize(14);
	spin->labelfont(FL_HELVETICA_BOLD);	
	spin->minimum(1);
	spin->maximum(250);
	spin->step(5);
	spin->value(hdr_step);	
	spin->align(Fl_Align(FL_ALIGN_LEFT));
	spin->callback(spinner_cb);
	
	apply = new Fl_Button(50,390,140,30,
								 "Redraw traces");
	apply->color(62);		apply->labelcolor(FL_BLACK);
	apply->labelfont(3);	apply->labelsize(16);	
	apply->callback(doRedraw);	
	
	
	add_wiggle = new Fl_Button(200, 393, 165, 25, 
									"Add/remove wiggles");
	add_wiggle->color(62);		add_wiggle->labelcolor(FL_BLACK);
	add_wiggle->labelfont(3);	add_wiggle->labelsize(14);										
	add_wiggle->callback(do_wiggle);
	
	////////////////////////////////////////////////////////////////////////
	///////////////////       S P E C T R U M         //////////////////////
	////////////////////////////////////////////////////////////////////////
	Fl_Box* label_ooo = new Fl_Box(FL_FRAME_BOX,20,430,380,24,
					"=======  Spectral Analysis of the displayed data  =======");
	label_ooo->labelfont(0);  label_ooo->labelcolor(7);
	label_ooo->color(36);		label_ooo->labelsize(14);
		
	Fl_Box* spec_tr = new Fl_Box(100,450,200,50,
									"Range of traces to accumulate spectrum :");
	spec_tr->labelfont(FL_HELVETICA_BOLD);
	spec_tr->labelsize(14);
	
	sp1_field = new Fl_Int_Input(70, 488, 50, 24, " from ");
	sprintf(cbuf,"%d",sp1);  sp1_field->value(cbuf);
	sp1_field->callback(input_sp1);
	
	sp2_field = new Fl_Int_Input(150, 488, 50, 24, " to ");
	sprintf(cbuf,"%d",sp2);  sp2_field->value(cbuf);
	sp2_field->callback(input_sp2);
	
	//////////   Callback function to open Spectral window    /////////////
	apply_spec = new Fl_Button(220, 485, 170, 30, 
									  " Calculate Spectrum ");
	apply_spec->color(62); apply_spec->labelsize(16); apply_spec->labelfont(3);
	apply_spec->labelcolor(FL_BLACK);          apply_spec->callback(Spectrum);
	
	
	Fl_Box* spec_gram = new Fl_Box(100,512,200,50,
						"Select a trace for Time-Frequency analysis :");
	spec_gram->labelfont(FL_HELVETICA_BOLD);
	spec_gram->labelsize(14);	
	
	trpick_field = new Fl_Int_Input(120, 551, 50, 24, " Trace # ");
	sprintf(cbuf,"%d",trpick);  trpick_field->value(cbuf);
	trpick_field->callback(input_trpick);
			
	//////////   Callback function to open SPECTROGRAM window    /////////////
	apply_spgr = new Fl_Button(190, 549, 200, 30, 
										"Calculate Spectrogram");
	apply_spgr->color(62); apply_spgr->labelsize(16); apply_spgr->labelfont(3);
	apply_spgr->labelcolor(FL_BLACK);          apply_spgr->callback(SPECgram);

	
	////////////////////////////////////////////////////////////////////////
	///////////////////      F I L T E R I N G        //////////////////////	
	////////////////////////////////////////////////////////////////////////
		
	Fl_Box* label_oo = new Fl_Box(FL_FRAME_BOX,20,590,380,24,
					"=======  Ormsby Band-Pass Filter Parameters  =======");
	label_oo->labelfont(0);  label_oo->labelcolor(7);
	label_oo->color(36);		label_oo->labelsize(14);
	
	Fl_Box* flt_param = new Fl_Box(80,615,260,50,
						"Specify four corner frequencies [ F1 - F4 ]\n\
of trapezoidal-shape frequency response:");
	flt_param->labelfont(FL_HELVETICA_BOLD);
	flt_param->labelsize(14);		
				
	
	f1_field = new Fl_Float_Input(250, 660, 50, 25, "0% Low Cut F1 = ");
	sprintf(cf1,"%5.1f",f1); f1_field->value(cf1);
	f1_field->labelsize(14); f1_field->labelfont(FL_HELVETICA_BOLD);
	f1_field->callback(input_f1);
		
	f2_field = new Fl_Float_Input(250, 690, 50, 25, "100% Low Cut F2 = ");
	sprintf(cf2,"%5.1f",f2); f2_field->value(cf2);
	f2_field->labelsize(14); f2_field->labelfont(FL_HELVETICA_BOLD);
	f2_field->callback(input_f2);
	
	f3_field = new Fl_Float_Input(250, 720, 50, 25, "100% High Cut F3 = ");
	sprintf(cf3,"%5.1f",f3); f3_field->value(cf3);
	f3_field->labelsize(14); f3_field->labelfont(FL_HELVETICA_BOLD);
	f3_field->callback(input_f3);
	
	f4_field = new Fl_Float_Input(250, 750, 50, 25, "0% High Cut F4 = ");
	sprintf(cf4,"%5.1f",f4); f4_field->value(cf4);
	f4_field->labelsize(14); f4_field->labelfont(FL_HELVETICA_BOLD);
	f4_field->callback(input_f4);			

	Fl_Box* comment2 = new Fl_Box(50,780,300,50,
							"The filter rejects below F1 and above F4,\n\
its response is flat from F2 to F3,\n\
 and has transitional slope from F1 to F2 and from F3 to F4");
	comment2->labelfont(FL_HELVETICA);
	comment2->labelsize(14);

	Fl_Input_Choice slope(270,840,100,30,"Select transitional slope shape: ");
	slope.callback(slope_choice, (void*)&slope);
	slope.labelsize(15); slope.labelfont(FL_HELVETICA_BOLD);	
	slope.add(" 1 - flat ");
	slope.add(" 2 - cosine");
	slope.value(1);  // cosine - default shape	
	
	//////////   Callback function to open Processed window    /////////////
	apply_flt = new Fl_Button(90, 880, 240, 30, 
									 " Apply Band-Pass Filter ");
	apply_flt->color(62); apply_flt->labelsize(16); apply_flt->labelfont(3);
	apply_flt->labelcolor(FL_BLACK);          apply_flt->callback(Process);	
	
	////////////////////////////////////////////////////////////////////////	
	////////////////////////////////////////////////////////////////////////		
	end_this = new Fl_Button(165,925, 80, 30, "Exit");
	end_this->color(88);
	end_this->labelcolor(FL_BLACK);
	end_this->labelfont(3);	
	end_this->labelsize(18);	
	end_this->callback(exit_b);	
			
	opts_window->end(); 
	opts_window->show();

	return(Fl::run());
	
	fclose(in_fp);
   return 0;
}
/////////////////////////////////////////////////////////////////////
/****    END_MAIN   END_MAIN   ENF_MAIN   END_MAIN   END_MAIN   ****/
/////////////////////////////////////////////////////////////////////
/****               Called Function Definition                  ****/
/////////////////////////////////////////////////////////////////////
void read_file_hdr(const char *name)
{
	int i=1,j,N,tr;
	long file_size,offset;
	float amp,count,avr,sinc;
	double val,std;
	char array[3601];
	short *hdr2,format;	
	char formats[][30] =    // string formats[9]
	{
	   "",
	   "4-byte IBM floating-point",     // 1
	   "4-byte integer",
	   "2-byte integer", 
	   "4-byte fixed-point with gain",
	   "4-byte IEEE floating-point",    // 5
	   "byte-swapped IEEE f.p.",        // 6
	   "non-standard 4-byte IEEE f.p.",  
	   "1-byte integer"
	};
	
	if( *( (char *)&i ) != 1 )	// TEST on COMPUTER ENDIANISM
	{
		fl_message("This application is supposed to run on\n\
Intel-based PC's with little-Endian data format.\n\
This is big-Endian computer. Close the window to exit");
		exit(1);
	}
	
	if ((in_fp = fopen(name, "rb")) == NULL)
	{fl_message("ERROR opening input file!\nClose the window to exit");exit(1);}
	
	fseek(in_fp, 0L, SEEK_END);	// go to file's end
	file_size = ftell(in_fp);
	fseek(in_fp, 0L, SEEK_SET);	// go to beginning	
	
   fread (&array, 3600, 1, in_fp);	// 1 object 3600 bytes long

	//   Convert Textual File Header:  EBCDIC-to-ASCII
	for(i=0;i<3200;i++) text_hdr[i]=to_ascii(array[i]);
	for(i=1;i<3200;i++) if(i%80==0) text_hdr[i-1]='\n';
		
   hdr2=(short*)array;
   format=hdr2[1612];			// Correspond to byte 3225
   
   if(format>0 && format<9) ;	// Native format - do nothing
   else							// swap bytes (for both IBM and IEEE)
   {
		for(i=0;i<24;i++) convert2(array+3212+i*2);
      format=hdr2[1612];		
		if(format>0 && format<9) ;
		else 
		{
			fl_message("ERROR: input is not a SEG-Y file!\nClose the window to exit");
			exit(1);
		}
   }   
	
   dsamp=(int)hdr2[1608];	// SAMPLE_INTERVAL at byte 3217 <-- GLOBAL
   nsamp=(int)hdr2[1610];	// NUMBER_SAMPLES at byte 3221  <-- GLOBAL
   measure=hdr2[1627];		// Measurement system at byte 3255
	
   if(dsamp<0 || dsamp>8000 || nsamp<10 || nsamp>16000)
   { 
		fl_message("ERROR:  irrational sampling information.\n\
Sample Interval = %d microseconds\nNumber of Trace Samples = %d\n\
Close the window to exit...",dsamp,nsamp);	
		fclose(in_fp);   exit(1);
   }

	tr_size = 60*sizeof(int) + nsamp*sizeof(float);

	Ntr = (int)((file_size-3600L)/(nsamp*sizeof(float)+240));
	
   convert_mode=1;  // No conversion
   if(format==5) convert_mode=2; // IEEE-to-native
   if(format==1) convert_mode=3;	// IBM--to-native
	
	///////////////////////////////////////////////////////////
	/****       Initial Visualization Parameters          ****/
	
	tr_sta=1;
	if(Ntr<50) tr_fin=Ntr;
	else tr_fin=50;

	tr_vis = tr_fin-tr_sta+1; // if trace count starts from '1'
	samp_sta=0;  samp_fin=10;
	if(nsamp<100) samp_fin=10;
	else if(nsamp<500) samp_fin=100;
			else if(nsamp<1002) samp_fin=500;
					else samp_fin=1000;				
	samp_vis = samp_fin-samp_sta;
	sp1=tr_sta;  sp2=tr_fin;
	gain = 1.0;
	
	/////////////////////////////////////////////////////////////
	/**      Get Normalization Factor for visualization       **/
	/**   by reading data until non-zero trace will be found  **/

	hdrs = new (nothrow) char[240];	
	seis = new (nothrow) float[nsamp];
	trace1 = new (nothrow) float[nsamp];	
	trace2 = new (nothrow) float[nsamp];
		
	Norm=1.0;
	for(tr=0;tr<Ntr;tr++)
	{
		offset=3600L+(long)tr_size*tr;	
		fseek(in_fp, offset, SEEK_SET);
		fread (hdrs, 240, 1, in_fp);
		if(convert_mode>1) convert_hdrs(hdrs);
		hdr2=(short*)hdrs; 
		if(hdr2[14]==2) continue;	// Trace ID code (0=Unknown,1=Seismic,2=Dead)
		fread (seis, nsamp*sizeof(float), 1, in_fp);
		if(convert_mode>1)
		{
			if(convert_mode==2)			// IEEE-to-native
				for(i=0;i<nsamp;i++) seis[i]=efloat(seis[i]);
			if(convert_mode==3)			// IBM-to-native
				for(i=0;i<nsamp;i++) seis[i]=ibm2ieee(seis[i]);			
		} 	
		
		count=avr=0.0;
		for(i=0;i<nsamp;i++) 
		{ 
			amp=fabs(seis[i]); 
			if(amp>0.0) { avr+=amp; count+=1.0; }
		}
		if(count>0.0) { avr/=count; Norm=4.0*avr; break; } 
	}
		
	if(hdr2[34] >= 0) elev_scal=(float)hdr2[34];
	else elev_scal=(float)(-1.0/hdr2[34]);
	if(hdr2[35] >= 0) coor_scal=(float)hdr2[35];
	else coor_scal=(float)(-1.0/hdr2[35]);	

//////////////////////////////////////////////////////////////////
//////////////   Info Screen Message   ///////////////////////////
   if(measure<1 || measure>2) strcpy(cgbuf,"unknown");
   if(measure==1) strcpy(cgbuf,"metric, meters");
   if(measure==2) strcpy(cgbuf,"imperial, feet");
	
	fl_message("File \"%s\"\nFile size = %ld kilobytes\nFile Header Info:\n\
======================================\n\
SEGY format code = %d ==> %s\n\
Sample Interval = %d microseconds\nNumber of Trace Samples = %d\n\
Measurement system = %d ==> %s\nNumber of traces = %d\n\
Average trace magnitude = %g\nElevation scalar value = %g\n\
Coordinate scalar value = %g\n\
======================================\n\
Close the window to continue...",
name,file_size/1000,format,formats[format],dsamp,nsamp,measure,\
cgbuf,Ntr,avr,elev_scal,coor_scal);
//////////////////////////////////////////////////////////////////

	/////////////      SPECTROGRAM      //////////////
		
	WIN = (int)(TWIN*1000/dsamp);	// Number of samples in 'TWIN'	
   N = WIN/2;
	SR = (int)(1000.01*1000/dsamp);	// Sample Rate, 1/sec
	fN=(float)SR/2.0;						// Nyquist frequency, Hz
	
	//////     Array of Gaussian [bell-shaped]  Weights	 	////////
	for(j=0;j<FOLD;j++) bell[j] =(double*)malloc(WIN*sizeof(double));
	//	Bell-shaped Gaussian = a*exp(-(x-b)*(x-b)/(2*std*std));
	// a = height of the bell's peak, b = position of the center
	// of the peak, and std = width of the 'bell' (standard deviation)
	switch(SR)
	{
		case 1000: sinc=8.0; break;
		case  500: sinc=4.0; break;
		case  250: sinc=2.0; break;
		default:   sinc=1.0; break;
	}
	for(j=0,std=sinc;j<FOLD;j++,std+=sinc)
	{
		norm[j]=0.0;
		for(i=0;i<WIN;i++)   // WIN - full window size [samples]
		{
			val = exp(-1.0*(N-i)*(N-i)/(2.0*std*std));
			bell[j][i] = val;		norm[j] += val;
		}
	}
}
////////////////////////////////////////////////////////////////
void read_seis_dat()		// fill up seis[] with data; float *seis;
{							// Do memory re-allocation before reading
	int i,tr;
	long offset;	
	
	delete[] hdrs; // originally allocated in read_file_hdr();	
	hdrs = new (nothrow) char[tr_vis * 240];		
	
	delete[] seis; // originally allocated in read_file_hdr();
	seis = new (nothrow) float[(tr_vis+1) * nsamp];
	if(seis == 0)
	{ 
		fl_message("ERROR: NO memory to accomodate seismic data!\n\
Close the window to exit...");	
		fclose(in_fp);  exit(1);
	}
	
/****   seis[..] contains the whole trace from t=0 to 'nsamp'   ****/

	for(tr=0;tr<tr_vis;tr++)
	{	// tr_size = 60*sizeof(int) + nsamp*sizeof(float);
		offset = 3600L + (long)tr_size*(tr_sta+tr-1);	
		fseek(in_fp, offset, SEEK_SET);
		fread (hdrs+tr*240, 240, 1, in_fp); // 240 bytes-long object
      if(convert_mode>1) convert_hdrs(hdrs+tr*240);		
		// fseek(in_fp, samp_sta*sizeof(float), SEEK_CUR);		
		fread (seis+tr*nsamp, nsamp*sizeof(float), 1, in_fp);
		if(convert_mode>1)
		{
			if(convert_mode==2)			// IEEE-to-native
				for(i=0;i<nsamp;i++) 
					seis[tr*nsamp+i]=efloat(seis[tr*nsamp+i]);
			if(convert_mode==3)			// IBM-to-native
				for(i=0;i<nsamp;i++) 
					seis[tr*nsamp+i]=ibm2ieee(seis[tr*nsamp+i]);			
		} 		
	}	
}
////////////////////////////////////////////////////////////////////
//	 Resampling using Cubic Convolution Method [R. Keys, 1981].
// Four data values from an array  dat[0..NS1]  are used every
// time to RETURN a calculated "index" value of the RESAMPLED
// array having size NS2 and same origin as the input array.
////////////////////////////////////////////////////////////////
float conv3(float *dat, int NS1, int NS2, int index)
{
	float y_1,y0,y1,y2,dx,dx2,dx3,k_1,k0,k1,k2;
	
	if(index<0 || index >=NS2) return 0.0;
	
	int i = index*NS1/NS2;
	
	y0=dat[i]; 	
	if(i<1) y_1=y0; else y_1=dat[i-1];
	if(i>NS1-2) y1=y0; else y1=dat[i+1]; 
	if(i>NS1-3) y2=y1; else y2=dat[i+2];
	
	dx = (float)(index*NS1 - i*NS2)/NS2;		
	dx2=dx*dx;  dx3=dx*dx2;		
	
	k_1 = (2.0*dx2 - dx3 - dx)/2.0;
	k0 = (3.0*dx3 - 5.0*dx2 + 2.0)/2.0;
	k1 = (4.0*dx2 - 3.0*dx3 + dx)/2.0;
	k2 = (dx3 - dx2)/2.0;
	return (k_1*y_1 + k0*y0 + k1*y1 + k2*y2);	
}
////////////////////////////////////////////////////////////////
void pal_interp(float palette[], int pal_size, float x, short colors[])
{  // based on "index_search" routine for linear interpolation
	int i,ilo,ihi;
	float dx,dr,dg,db,r0,g0,b0,rf,gf,bf;	
	
	ilo = 0;
	ihi = pal_size-1;
	
	while(ihi > ilo + 1) 
	{
		i = (ihi + ilo)/2;
		if(palette[i] > x) ihi = i;
		else ilo = i;
	}
	
	dx = (x-palette[ilo])/(palette[ilo+1]-palette[ilo]);
	r0 = palette[pal_size+ilo];
	g0 = palette[pal_size*2+ilo];
	b0 = palette[pal_size*3+ilo];
	dr = palette[pal_size+ilo+1] - r0;
	dg = palette[pal_size*2+ilo+1] - g0;
	db = palette[pal_size*3+ilo+1] - b0;
	
	rf = r0 + dx*dr; if(rf<0.1) rf =0.1; if(rf>255.1) rf =255.1;
	gf = g0 + dx*dg; if(gf<0.1) gf =0.1; if(gf>255.1) gf =255.1;
	bf = b0 + dx*db; if(bf<0.1) bf =0.1; if(bf>255.1) bf =255.1;
	
	colors[0] = (short)rf;
	colors[1] = (short)gf;	
	colors[2] = (short)bf;	
}
//
void convert_hdrs(char* array)  //  Byte-swapping for the
{                               //  240-byte array of trace headers
   int k;
   void convert2(char* input);
   void convert4(char* input);
	
   for(k=0;k<7;k++) convert4(array+k*4);       // Bytes   1 - 28
   for(k=0;k<4;k++) convert2(array+28+k*2);    // Bytes  29 - 36
   for(k=0;k<8;k++) convert4(array+36+k*4);    // Bytes  37 - 68
   for(k=0;k<2;k++) convert2(array+68+k*2);    // Bytes  69 - 72
   for(k=0;k<4;k++) convert4(array+72+k*4);    // Bytes  73 - 88
   for(k=0;k<46;k++) convert2(array+88+k*2);   // Bytes  89 - 180
   for(k=0;k<5;k++) convert4(array+180+k*4);   // Bytes 181 - 200
   for(k=0;k<2;k++) convert2(array+200+k*2);   // Bytes 201 - 204
   convert4(array+204);  convert2(array+208);  // Bytes 205 - 210
   for(k=0;k<5;k++) convert2(array+210+k*2);   // Bytes 211 - 224
   convert4(array+224);  convert2(array+228);  // Bytes 225 - 230
   convert2(array+230);                        // Bytes 231 - 233
	// Unassigned bytes 233 - 240
}
////////////////////////////////////////////////////////////////
//
// SUBROUTINES
// WRITTEN BY MARK GOLDMAN
// U.S. GEOLOGICAL SURVEY
// http://pubs.usgs.gov/of/2005/1311/
//
// Adopted by Yuri Ganshin
//
void convert2(char* input) {
	
   unsigned char temp;
	
   temp = input[1];
   input[1] = input[0];
   input[0] = temp;
}

void convert4(char* input) {
	
   unsigned char temp;
	
   temp = input[3];
   input[3] = input[0];
   input[0] = temp;
	
   temp = input[2];
   input[2] = input[1];
   input[1] = temp;
}

float efloat(float value) {
	
   union {
      char bytes[4];
      float flt;
   } ieee;
	
	// swap bytes
   unsigned char temp;
   ieee.flt = value;
	
   temp = ieee.bytes[3];
   ieee.bytes[3] = ieee.bytes[0];
   ieee.bytes[0] = temp;
	
   temp = ieee.bytes[2];
   ieee.bytes[2] = ieee.bytes[1];
   ieee.bytes[1] = temp;

   return(ieee.flt);
}

float ibm2ieee(const float ibm) 
{
   long IntMantissa;
   float Mantissa,PosResult;
   // DEFINE AND SETUP UNION VARIABLES
   union { float f; unsigned char c[4]; } src; 
	
   // ASSIGN INPUT PARAMETER TO UNION MEMBER
   src.f = ibm; // When you change the value of a union member, 
	// you modify all the members of the union
   // CONVERTION
   IntMantissa = ((long)src.c[1] << 16)+((long)src.c[2]<<8)+src.c[3];
   Mantissa = (float)(IntMantissa)/(float)(0x1000000);
   PosResult = Mantissa*pow(16.0,(double)((src.c[0]&0x7F)-64));
	
   if (src.c[0] & 0x80) return -PosResult;
   else return PosResult;
}

/*
 **  ASCII <=> EBCDIC conversion functions
 */
unsigned char to_ascii ( unsigned char c )
{
	unsigned char e2a[256] = {
		0,  1,  2,  3,156,  9,134,127,151,141,142, 11, 12, 13, 14, 15,
		16, 17, 18, 19,157,133,  8,135, 24, 25,146,143, 28, 29, 30, 31,
		128,129,130,131,132, 10, 23, 27,136,137,138,139,140,  5,  6,  7,
		144,145, 22,147,148,149,150,  4,152,153,154,155, 20, 21,158, 26,
		32,160,161,162,163,164,165,166,167,168, 91, 46, 60, 40, 43, 33,
		38,169,170,171,172,173,174,175,176,177, 93, 36, 42, 41, 59, 94,
		45, 47,178,179,180,181,182,183,184,185,124, 44, 37, 95, 62, 63,
		186,187,188,189,190,191,192,193,194, 96, 58, 35, 64, 39, 61, 34,
		195, 97, 98, 99,100,101,102,103,104,105,196,197,198,199,200,201,
		202,106,107,108,109,110,111,112,113,114,203,204,205,206,207,208,
		209,126,115,116,117,118,119,120,121,122,210,211,212,213,214,215,
		216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,
		123, 65, 66, 67, 68, 69, 70, 71, 72, 73,232,233,234,235,236,237,
		125, 74, 75, 76, 77, 78, 79, 80, 81, 82,238,239,240,241,242,243,
		92,159, 83, 84, 85, 86, 87, 88, 89, 90,244,245,246,247,248,249,
		48, 49, 50, 51, 52, 53, 54, 55, 56, 57,250,251,252,253,254,255
	};
	
	return e2a[c];
}
